//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: 21 December 1997
// Last Modified: Wed Feb 11 22:43:15 GMT-0800 1998
// Filename:      ...sig/code/control/RadioDrum/RadioDrum.cpp
// Syntax:        C++
// $Smake:        cc -Wall -g -c %b.cc -I../../../include && rm -f %b.o
//
// Description: A structure for handling the RadioDrum variables.
//


#include <ctype.h>
#include <iomanip.h>

#ifdef LINUX
   #include <unistd.h>
#endif
#ifdef SUN
   #include <unistd.h>
#endif
#ifdef IRIX
   #include <unistd.h>
#endif

#ifndef VISUAL
   #include <libc.h>
#endif


// Below are the global state variables for the baton which
// get updated as soon as a message command is received and
// before a complete set of trigger or position data is received.
//
// With these variables will not work if there are more than
// one Radio Drum.
//

// instantaneous position variables:
short pt1, pt2, pt3, pt4;             // current pot positions
short xt1, yt1, zt1, xt2, yt2, zt2;   // current stick positions

// trigger variables:
short xw1, yw1, xw2, yw2;             // whack positions of batons at trigger
short wh1, wh2;                       // whack strength of a trigger
short whack1, whack2;                 // booleans for triggers occuring
short whack1x, whack2x;               // booleans for triggers occuring
short whack1y, whack2y;               // booleans for triggers occuring
long  trigtime1, trigtime2;           // times that the sticks were triggered

// trigger plane variables:
short hit1, set1, hit2, set2;         // first four elements of calib array


// Later the functions below will go into a behavior class which 
// will be inherited by the RadioBaton class.  These functions
// are defined by the user who creates an improv program.

extern void stick1trig(void);
extern void stick2trig(void);
extern void b14plustrig(void);
extern void b15plustrig(void);
extern void b14minusuptrig(void);
extern void b14minusdowntrig(void);
extern void b15minusuptrig(void);
extern void b15minusdowntrig(void);


// state tags for recording radio baton variables:
#define TRIGGER1RECORD     "t1"
#define TRIGGER2RECORD     "t2"
#define POSITION1RECORD    "p1"
#define POSITION2RECORD    "p2"
#define DIAL1RECORD        "d1"
#define DIAL2RECORD        "d2"
#define DIAL3RECORD        "d3"
#define DIAL4RECORD        "d4"
#define FOOTPEDAL1RECORD   "f1"
#define FOOTPEDAL2RECORD   "f2"
#define BUTTON1RECORD      "b1"
#define BUTTON2RECORD      "b2"



///////////////////////////////////////////////////////////////////////////


#include <fstream.h>
#include "RadioDrum.h"


//////////////////////////////
// 
// RadioDrum::RadioDrum
//

RadioDrum::RadioDrum(void) : MidiIO() {
   setStateSize(DEFAULT_STATE_SIZE);
   for (int i=0; i<16; i++) {
      completeBufQ[i] = 0;
   }
   for (int j=0; j<22; j++) {
      calib[j] = 0;
   }
   reportingQ = 0;
   recordStateQ = 0;
   storedReportStatus = 0;
   errorQ = 0;
   writingCalib = 0;
   calibFilename = NULL;
   calibFilename = new char[100];
   strcpy(calibFilename, "calib.def");
}


RadioDrum::RadioDrum(int outputDevice, int inputDevice) :
      MidiIO(outputDevice, inputDevice) {
   setStateSize(DEFAULT_STATE_SIZE);
   for (int i=0; i< 16; i++) {
      completeBufQ[i] = 0;
   }
   for (int j=0; j<22; j++) {
      calib[j] = 0;
   }
   reportingQ = 0;
   recordStateQ = 0;
   storedReportStatus = 0;
   errorQ = 0;
   writingCalib = 0;
   calibFilename = NULL;
   calibFilename = new char[100];
   strcpy(calibFilename, "calib.def");
}



//////////////////////////////
//
// RadioDrum::~RadioDrum
//

RadioDrum::~RadioDrum() {
   if (calibFilename != NULL) {
      delete [] calibFilename;
      calibFilename = NULL;
   }
   if (recordStateQ == 1) {
      recordStateStop();
      recordStateQ = 0;
   }
}



//////////////////////////////
//
// RadioDrum::centerStick1
//

void RadioDrum::centerStick1(void) {
   sendMessage(SET_CENTER_STICK1);
}



//////////////////////////////
//
// RadioDrum::centerStick2
//

void RadioDrum::centerStick2(void) {
   sendMessage(SET_CENTER_STICK2);
}



//////////////////////////////
//
// RadioDrum::decreaseXYSensitivity
//
 
void RadioDrum::decreaseXYSensitivity(void) {
   sendMessage(DEC_XY_SENSITIVITY);
}



//////////////////////////////
//
// RadioDrum::decreaseZSensitivity
//

void RadioDrum::decreaseZSensitivity(void) {
   sendMessage(DEC_Z_SENSITIVITY);
}



//////////////////////////////
//
// RadioDrum::getBuf -- returns the currently
//    stored value for the given buffer.  Returns
//    -1 if the radio baton has not completed sending the 
//    value data, which is a 12-bit unsigned value
//    sent in 4 nibbles (most significant nibble is always 0).
//    Will also return -1 if an invalid array location is
//    accessed.
//

int RadioDrum::getBuf(int index) {
   if (index < 0 || index > 15) {
      cerr << "Invalid array index for buf: " << index << endl;
      return -1;
   }

   if (completeBufQ[index] == 0) {
      return -1;
   } else {
      return buf[index];
   }
}



//////////////////////////////
//
// RadioDrum::getError -- return the state of the 
//    MIDI input error printing value
//

int RadioDrum::getError(void) const {
   return errorQ;
}



//////////////////////////////
//
// RadioDrum::getPositionReporting
//

int RadioDrum::getPositionReporting(void) const {
   return reportingQ;
}



//////////////////////////////
//
// RadioDrum::getReportStatus
//

int RadioDrum::getReportStatus(void) const {
   return reportingQ;
}



//////////////////////////////
//
// RadioDrum::getWhack1Time -- returns the last trigger time of
//    stick 1.
//

long RadioDrum::getWhack1Time(void) const {
   return t1t_temp;
}



//////////////////////////////
//
// RadioDrum::getWhack2Time -- returns the last trigger time of
//    stick 2
//

long RadioDrum::getWhack2Time(void) const {
   return t2t_temp;
}



//////////////////////////////
//
// RadioDrum::increaseXYSensitivity(void) {
//
 
void RadioDrum::increaseXYSensitivity(void) {
   sendMessage(INC_XY_SENSITIVITY);
}



//////////////////////////////
//
// RadioDrum::increaseZSensitivity
//

void RadioDrum::increaseZSensitivity(void) {
   sendMessage(INC_Z_SENSITIVITY);
}



//////////////////////////////
//
// RadioDrum::positionReportingOff
//

int RadioDrum::positionReportingOff(void) {
   int oldState = reportingQ;
   rawsend(BAT_MIDI_COMMAND, REPORT_POS, POS_REPORT_OFF);
   reportingQ = 0;
   return oldState;
}



//////////////////////////////
// 
// RadioDrum::positionReportingOn
//

int RadioDrum::positionReportingOn(void) { 
   int oldState = reportingQ;
   rawsend(BAT_MIDI_COMMAND, REPORT_POS, POS_REPORT_ON);
   reportingQ = 1;
   return oldState;
}



//////////////////////////////
//
// RadioDrum::processIncomingMessages
// 

void RadioDrum::processIncomingMessages(void) {
   while (MidiInput::getCount() > 0) {
      interpretCommand(MidiInput::extract());
   }
}



//////////////////////////////
//
// RadioDrum::readCalibrationFile -- send calibration data from
//    radio baton to the specified file if file couldn't be
//    found, return 0; otherwise, returns 1.
//

int RadioDrum::readCalibrationFile(const char* aFilename) {
   ifstream input(aFilename, ios::nocreate);
   if (!input) {
      return 0;
   }
   char tempString[256];
   
   input >> tempString;     // hit1
   input >> calib[0];
   input >> tempString;     // set1
   input >> calib[1];
   input >> tempString;     // hit2
   input >> calib[2];
   input >> tempString;     // set2
   input >> calib[3];
   input >> tempString;     // zero1x
   input >> calib[4];
   input >> tempString;     // zero1y
   input >> calib[5];
   input >> tempString;     // zero2x
   input >> calib[6];       
   input >> tempString;     // zero2y
   input >> calib[7];
   input >> tempString;     // range
   input >> calib[8];       
   input >> tempString;     // restpoint
   input >> calib[9];
   for (int i=10; i<21; i++) {
      input >> calib[i];
   }
   input >> tempString;      // zcalib
   input >> calib[21];
 
   // send calibration data to the radio drum
   // only if calibration file is read completly
   for (int j=0; j<6; j++) {
      if (tempString[j] == '\0') break;
      tempString[j] = tolower(tempString[j]);
   }
   if (strcmp("zcalib", tempString) == 0) {
      // file was read successfully,
      // update the global state varaibles: hit and set levels of sticks
      hit1 = calib[0];
      set1 = calib[1];
      hit2 = calib[2];
      set2 = calib[3];

      // now tell the radio drum what the calibration is:
      sendCalibration();      

      return 1;
   } else {
      return 0;
   }
}



//////////////////////////////
//
// RadioDrum::recordingQ -- returns true if currently recording,
//	false if not recording.
//

int RadioDrum::recordingQ(void) const {
   return recordStateQ;
}



//////////////////////////////
//
// RadioDrum::recordStateStart -- 
//

void RadioDrum::recordStateStart(const char* aFileName) {
   if (recordStateQ == 1) {
      recordStateStop();
   }

   recordOutput.open(aFileName, ios::out);
   if (!recordOutput) {   // open file failed
      cerr << "Error: cannot open file " << aFileName << endl;
      recordStateQ = 0;
   } else {
      recordStateQ = 1;
      timer.reset();
      recordTimeOffset = -1;
   }
}



//////////////////////////////
//
// RadioDrum::recordStateStop
//

void RadioDrum::recordStateStop(void) {
   if (recordStateQ == 0)  return;
   if (recordOutput.is_open()) {
      recordOutput.close();
      recordTimeOffset = -1;
      recordStateQ = 0;
   }
}



//////////////////////////////
//
// RadioDrum::requestCalibration -- Send a request for calibration data 
//    from the radio drum.  It will be received later from the 
//    radio drum and stored in the calib array.
//

void RadioDrum::requestCalibration(void) {
   MidiOutput::rawsend(BAT_MIDI_COMMAND, REPORT_CALIB, 0);
}



//////////////////////////////
//
// RadioDrum::requestBufData -- Gets all 16 values of buf data
//    from the radio drum.
//

void RadioDrum::requestBufData(void) {
   // send request for buffer data
   for (int i=0; i<16; i++) {
      MidiOutput::rawsend(BAT_MIDI_COMMAND, REPORT_BUF, i);
   }
   // the processIncomingMessages function will detect the results
   // of this buff data request and fill the buff array.
}



//////////////////////////////
//
// RadioDrum::sendMessage -- sends a message to the radio drum.
//    Messages are those related to the BAT_SEND set
//    found in baton.h .
//

void RadioDrum::sendMessage(int aMessage) {
   if (aMessage == TEST_BATON) {
      testBaton();
   } else {
      rawsend(BAT_MIDI_COMMAND, BAT_SEND, aMessage);
   }
}



//////////////////////////////
//
// RadioDrum::setError -- if true, then will print out
//    unknown MIDI commands which seem to be bad.
//    In other words, if an 0xa0 command byte was 
//    received, and it doesn't match any baton messages,
//    then an error message will be print to cerr.
//    If false, then don't print any message for bad
//    commands.
//

void RadioDrum::setError(int errorStatus) {
   if (errorStatus) {
      errorQ = 1;
   } else {
      errorQ = 0;
   }
}



//////////////////////////////
//
// RadioDrum::setReportStatus
//

void RadioDrum::setReportStatus(int aStatus) {
   if (aStatus == 0) {
      reportingQ = 0;
      positionReportingOff();
   } else {
      reportingQ = 1;
      positionReportingOn();
   }
}


//////////////////////////////
//
// RadioDrum::setStateSize -- set the length of the
//    histories of each state variable in the object.
//

void RadioDrum::setStateSize(int aSize) {
   if (aSize <= 0) {
      cerr << "Error: state size too small: " << aSize << endl;
      exit(1);
   }
   x1p.setSize(aSize);   // stick1 x-axis position
   y1p.setSize(aSize);   // stick1 y-axis position
   z1p.setSize(aSize);   // stick1 z-axis position
   x2p.setSize(aSize);   // stick2 x-axis position
   y2p.setSize(aSize);   // stick2 y-axis position
   z2p.setSize(aSize);   // stick2 z-axis position   
   d1p.setSize(aSize);   // dial 1 position
   d2p.setSize(aSize);   // dial 2 position
   d3p.setSize(aSize);   // dial 3 position
   d4p.setSize(aSize);   // dial 4 position
   t1t.setSize(aSize);   // trigger stick 1 time
   x1t.setSize(aSize);   // stick1 x-axis trigger pos.
   y1t.setSize(aSize);   // stick1 y-axis trigger pos.
   w1t.setSize(aSize);   // stick1 wack at trigger time
   t2t.setSize(aSize);   // trigger stick 2 time
   x2t.setSize(aSize);   // stick2 x-axis trigger pos.
   y2t.setSize(aSize);   // stick2 y-axis trigger pos.
   w2t.setSize(aSize);   // stick2 whach at trigger time
}



//////////////////////////////
//
// RadioDrum::setStickLevels
//

void RadioDrum::setStickLevels(void) {
   sendMessage(SET_STICK_LEVELS);
}



//////////////////////////////
//
// RadioDrum::testBaton -- checks to see if radio drum is connected
//

int RadioDrum::testBaton(void) {
   int reportStatus = positionReportingOff();
   rawsend(BAT_MIDI_COMMAND, BAT_SEND, TEST_BATON);

   #ifdef VISUAL
      Sleep(20);           // 20 milliseconds
   #else
      usleep(20000);       // 20000 microseconds = 20 milliseconds
   #endif

   // search for OK receive message
   if (MidiInput::getCount() == 0) {
      setReportStatus(reportStatus);
      return 0;
   } else {
      for (int i=0; i<MidiInput::getCount(); i++) {
         if ( (*this)[i].command() == BAT_MIDI_COMMAND &&
              (*this)[i].p1() == BAT_SEND &&
              (*this)[i].p2() == 0 ) {
                  setReportStatus(reportStatus);
                  return 1;
         }
      }
   }
   setReportStatus(reportStatus);
   return 0;
}



//////////////////////////////
//
// RadioDrum::toggleError -- switches the printing error
//    status when the class received a MIDI messages it
//    doesn't understand (and probably shouldn't have
//    ignored.)
//

void RadioDrum::toggleError(void) {
   errorQ = !errorQ;
}



//////////////////////////////
//
// RadioDrum::togglePositionReporting
//

void RadioDrum::togglePositionReporting(void) {
   if (reportingQ) {
      reportingQ = 0;
      positionReportingOff();
   } else {
      reportingQ = 1;
      positionReportingOn();
   }
}



//////////////////////////////
//
// RadioDrum::writeCalibrationFile -- gets the calibration
//    data from the radio drum and writes it to the 
//    specified file.
//

void RadioDrum::writeCalibrationFile(const char* aFilename) {
   pushReportingStatus();
   positionReportingOff();

   // ask for calibration data from the radio drum   
   requestCalibration();
   writingCalib = 1;

   // store the filename for writeCalibrationData() function
   if (calibFilename != NULL) {
      delete [] calibFilename;
   }
   calibFilename = new char[strlen(aFilename) + 1];
   strcpy(calibFilename, aFilename);
}





///////////////////////////////////////////////////////////////////////////
//
// private functions
//


//////////////////////////////
//
// RadioDrum::interpretCommand
//

void RadioDrum::interpretCommand(MidiMessage aMessage) {
   static ushort value;  // for the buff value receive commands

   if (aMessage.command() == BAT_MIDI_COMMAND) {
      switch (aMessage.p1()) {
         case BAT_STICK1_RESPONSE_X:         // stick 1 responding to poll; x
            s1ps(aMessage.time);
            s1pd(DATA_X, aMessage.p2());
            break;
         case BAT_STICK1_RESPONSE_Y:         // stick 1 responding to poll; y
            s1pd(DATA_Y, aMessage.p2());
            break;
         case BAT_STICK1_RESPONSE_Z:         // stick 1 responding to poll; z
            s1pd(DATA_Z, aMessage.p2());
            break;
         case BAT_STICK2_RESPONSE_X:         // stick 2 responding to poll; x
            s2ps(aMessage.time);
            s2pd(DATA_X, aMessage.p2());
            break;
         case BAT_STICK2_RESPONSE_Y:         // stick 2 responding to poll; y
            s2pd(DATA_Y, aMessage.p2());
            break;
         case BAT_STICK2_RESPONSE_Z:         // stick 2 responding to poll; z
            s2pd(DATA_Z, aMessage.p2());
            break;
         case BAT_POT1_RESPONSE:             // pot 1 responding to poll
            pt1 = aMessage.p2();             // update global state variable
            recordState(aMessage.time, DIAL1RECORD, pt1);
            d1p.insert(aMessage.p2());
            break;
         case BAT_POT2_RESPONSE:             // pot 2 responding to poll
            pt2 = aMessage.p2();             // update global state variable
            recordState(aMessage.time, DIAL2RECORD, pt2);
            d2p.insert(aMessage.p2());
            break;
         case BAT_POT3_RESPONSE:             // pot 3 responding to poll
            pt3 = aMessage.p2();             // update global state variable
            recordState(aMessage.time, DIAL3RECORD, pt3);
            d3p.insert(aMessage.p2());
            break;
         case BAT_POT4_RESPONSE:             // pot 4 responding to poll
            pt4 = aMessage.p2();             // update global state variable
            recordState(aMessage.time, DIAL4RECORD, pt4);
            d4p.insert(aMessage.p2());
            break;
         case BAT_STICK1_TRIGGER:            // stick 1 got triggered
            s1ts(aMessage.time);
            s1td(DATA_W, aMessage.p2());
            break;
         case BAT_STICK1_TRIG_X:             // stick 1 got triggered
            s1td(DATA_X, aMessage.p2());
            break;
         case BAT_STICK1_TRIG_Y:             // stick 1 got triggered
            s1td(DATA_Y, aMessage.p2());
            break;
         case BAT_STICK2_TRIGGER:            // stick 2 got triggered
            s2ts(aMessage.time);
            s2td(DATA_W, aMessage.p2());
            break;
         case BAT_STICK2_TRIG_X:             // stick 2 got triggered
            s2td(DATA_X, aMessage.p2());
            break;
         case BAT_STICK2_TRIG_Y:             // stick 2 got triggered
            s2td(DATA_Y, aMessage.p2());
            break;

         case BAT_BUTTON_FOOT_TRIGGER:       // button or pedal was triggered
            switch (aMessage.p2()) {
               case BAT_B14p_TRIGGER:        // B14+ button pressed
                  b14pt = aMessage.time;
                  recordState(aMessage.time, BUTTON1RECORD);
                  b14plustrig();
                  break;
               case BAT_B15p_TRIGGER:        // B15+ button pressed
                  b15pt = aMessage.time;
                  recordState(aMessage.time, BUTTON2RECORD);
                  b15plustrig();
                  break;
               case BAT_B14m_DOWN_TRIGGER:   // B14- pedal was depressed
                  b14mdt = aMessage.time;
                  recordState(b14mdt, FOOTPEDAL1RECORD, 1);
                  b14minusdowntrig();
                  break;
               case BAT_B14m_UP_TRIGGER:     // B14- pedal was released
                  b14mut = aMessage.time;
                  recordState(b14mut, FOOTPEDAL1RECORD, 0);
                  b14minusuptrig();
                  break;
               case BAT_B15m_DOWN_TRIGGER:   // B15- pedal was depressed
                  b15mdt = aMessage.time;
                  recordState(b15mut, FOOTPEDAL2RECORD, 1);
                  b15minusdowntrig();
                  break;
               case BAT_B15m_UP_TRIGGER:     // B15- pedal was released
                  b15mut = aMessage.time;
                  recordState(b15mut, FOOTPEDAL2RECORD, 0);
                  b15minusuptrig();
                  break;
            }
            break;

         case BAT_SEND:                     // baton alive message et al.
            // ignore this command it is handled in the testBaton function
            break;   
         default:
            if (errorQ) {
               cerr << "Baton command not recognized: 0x" << hex
                    << (int)aMessage.command() << dec
                    << ", param1 = " << (int)aMessage.p1()
                    << ", param2 = " << (int)aMessage.p2()
                    << endl;
            }
      }  // end of switch (aMessage.command())

   }  // end of if a baton command (0xa0)

   // take care of calibration data
   // nibbles are replaced as they come in from the radio drum
   else if (aMessage.command() == 0xa1) {
      // if p1 is not in the range from 0 to 21 then ignore the error:
      if (aMessage.p1() > 21) return;
      value = (ushort)aMessage.p2();
      value &= 0x000f;                         // guarentee no high bits
      calib[aMessage.p1()] &= 0x0fff;          // blank out old nibble
      calib[aMessage.p1()] |= value << 12;
   } else if (aMessage.command() == 0xa2) {
      // if p1 is not in the range from 0 to 21 then ignore the error:
      if (aMessage.p1() > 21) return;
      value = (ushort)aMessage.p2();
      value &= 0x000f;                         // guarentee no high bits
      calib[aMessage.p1()] &= 0xf0ff;          // blank out old nibble
      calib[aMessage.p1()] |= value << 8;
   } else if (aMessage.command() == 0xa3) {
      // if p1 is not in the range from 0 to 21 then ignore the error:
      if (aMessage.p1() > 21) return;
      value = (ushort)aMessage.p2();
      value &= 0x000f;                         // guarentee no high bits
      calib[aMessage.p1()] &= 0xff0f;          // blank out old nibble
      calib[aMessage.p1()] |= value << 4;
   } else if (aMessage.command() == 0xa4) {
      // if p1 is not in the range from 0 to 21 then ignore the error:
      if (aMessage.p1() > 21) return;
      value = (ushort)aMessage.p2();
      value &= 0x000f;                         // guarentee no high bits
      calib[aMessage.p1()] &= 0xfff0;          // blank out old nibble
      calib[aMessage.p1()] |= value;
 
      // write the calib file if last nibble of last element received:
      if (aMessage.p1() == 21 && writingCalib) {
         writingCalib = 0;

         // finished receiving calib data from radio drum.  The data
         // is sent only when a user request to write data to a file.
	 // so write the file now.
         writeCalibrationData();
         cerr << "Calibration data written to file: "
              << calibFilename << endl;

         // set the position reporting to what it was before the
         // calibration was requested from the radio drum
         popReportingStatus();
      
         // update the global variables which rely on the calib array:
         hit1 = calib[0];
         set1 = calib[1];
         hit2 = calib[2];
         set2 = calib[3];
      }
   }

   // take care of buf data
   else if (aMessage.command() == 0xa5) {
      // if p1 is not in the range from 0 to 15 then ignore the error:
      if (aMessage.p1() > 15) return;
      completeBufQ[aMessage.p1()] = 0;
      value = aMessage.p2();
      buf[aMessage.p1()] = (value << 12);
   } else if (aMessage.command() == 0xa6) {
      // if p1 is not in the range from 0 to 15 then ignore the error:
      if (aMessage.p1() > 15) return;
      value = aMessage.p2();
      buf[aMessage.p1()] += (value << 8);
   } else if (aMessage.command() == 0xa7) {
      // if p1 is not in the range from 0 to 15 then ignore the error:
      if (aMessage.p1() > 15) return;
      value = aMessage.p2();
      buf[aMessage.p1()] += (value << 4);
   } else if (aMessage.command() == 0xa8) {
      // if p1 is not in the range from 0 to 15 then ignore the error:
      if (aMessage.p1() > 15) return;
      buf[aMessage.p1()] += aMessage.p2();
      completeBufQ[aMessage.p1()] = 1;
   }

   // All other MIDI commands are ignored.  Maybe they are
   // for/from a synthesizer or something like that.

}



//////////////////////////////
//
// RadioDrum::pushReportingStatus
//

void RadioDrum::pushReportingStatus(void) {
   storedReportStatus = reportingQ;
}



//////////////////////////////
//
// RadioDrum::popReportingStatus
//
 
void RadioDrum::popReportingStatus(void) {
   if (storedReportStatus == reportingQ)  return;
   else togglePositionReporting();
}



//////////////////////////////
//
// RadioDrum::recordState
//

void RadioDrum::recordState(long aTime, const char* aState) {
   if (recordingQ()) {
      if (recordTimeOffset == -1) {
         recordTimeOffset = aTime;
      }
      recordOutput << setw(6) << aTime-recordTimeOffset 
                   << ' ' << aState << '\n';
   }
}


void RadioDrum::recordState(long aTime, const char* aState, int value) {
   if (recordingQ()) {
      if (recordTimeOffset == -1) {
         recordTimeOffset = aTime;
      }
      recordOutput << setw(6) << aTime - recordTimeOffset << ' ' 
                   << aState << setw(3) << value << '\n';
   }
}


void RadioDrum::recordState(long aTime, const char* aState, 
      int value1, int value2, int value3) {
   if (recordingQ()) {
      if (recordTimeOffset == -1) {
         recordTimeOffset = aTime;
      }
      recordOutput << setw(6) << aTime - recordTimeOffset << ' ' << aState 
                   << setw(3) << value1 
                   << setw(3) << value2 
                   << setw(3) << value3 << '\n';
   }
}



//////////////////////////////
//
// RadioDrum::sendCalibration -- Sends the 
//    calibration data stored in object to the
//    radio drum.  size of calib array is 22.
//

void RadioDrum::sendCalibration(void) {
   for (int j=0; j<22; j++) {
      MidiOutput::rawsend(0xa1, j, (calib[j] >> 12) & 0x000f);
      MidiOutput::rawsend(0xa2, j, (calib[j] >> 8)  & 0x000f);
      MidiOutput::rawsend(0xa3, j, (calib[j] >> 4)  & 0x000f);
      MidiOutput::rawsend(0xa4, j, (calib[j])       & 0x000f);
   }
}



//////////////////////////////
//
// RadioDrum::s1td -- "Stick 1 Trigger Data"
//    Stores trigger position data in temporary storage
//    until a complete set of tigger data is received.
//

void RadioDrum::s1td(int flag, uchar aValue) {
   switch (flag) {
      case DATA_W:
         wh1 = aValue;         // set the global state variable
                               // whack1, trigtime1 set in the function s1ts
         w1t_temp = aValue;    // set the internal temporary value
         s1tf += 1;            // set flag bit 1 (or create an error)
         break;
      case DATA_X:
         xw1 = aValue;         // set the global state variable
         whack1x = 1;          // set the global state variable
         x1t_temp = aValue;    // set the internal temporary value
         s1tf += 4;            // set flag bit 3 (or create an error)
         break;
      case DATA_Y:
         yw1 = aValue;         // set the global state variable
         whack1y = 1;          // set the global state variable
         y1t_temp = aValue;    // set the internal temporary value
         s1tf += 16;           // set flag bit 5 (or create an error)
         if (s1tf == 21) {     // if complete data then save trigger set
            t1t.insert(t1t_temp);
            w1t.insert(w1t_temp);
            x1t.insert(x1t_temp);
            y1t.insert(y1t_temp);
            recordState(t1t_temp, TRIGGER1RECORD, 
               x1t_temp, y1t_temp, w1t_temp);
         } else {
//            cerr << "Error: incomplete stick 1 position information" 
//                 << ", flag value is: " << (int)s1tf << endl;
         }
         break;
   }
}



//////////////////////////////
//
// RadioDrum::s1ts -- "Stick 1 Trigger Status"
//    Keeps track of a compete set of stick 1 trigger positions.  
//    This function resets temporary values to a new set of 
//    stick 1 trigger values.   Stores the timestamp and initilaizes
//    the s1tf status flag.
//

void RadioDrum::s1ts(long aTime) {
   whack1 = 1;                 // set the global state variable
   trigtime1 = aTime;          // set the global state variable 
   t1t_temp = aTime;           // set the internal temporary value
   s1tf = 0x00;                // initialize the flag to determine good set
   stick1trig();               // call the user state function
}



//////////////////////////////
//
// RadioDrum::s2td -- "Stick 2 Trigger Data"
//    Stores trigger position data in temporary storage
//    until a complete set of tigger data is received.
//

void RadioDrum::s2td(int flag, uchar aValue) {
   switch (flag) {
      case DATA_W:
         wh2 = aValue;          // set the global state variable
                                // whack2, trigtime2 set in s2ts function
         w2t_temp = aValue;     // set the internal temporary value
         s2tf += 1;             // set flag bit 1 (or create an error)
         break;
      case DATA_X:
         xw2 = aValue;          // set the global state variable
         whack2x = 1;           // set the global state variable
         x2t_temp = aValue;     // set the internal temporary value
         s2tf += 4;             // set flag bit 3 (or create an error)
         break;
      case DATA_Y:
         yw2 = aValue;          // set the global state variable
         whack2y = 1;           // set the global state variable
         y2t_temp = aValue;     // set the internal temporary value
         s2tf += 16;            // set flag bit 5 (or create an error)
         if (s2tf == 21) {      // if complete trigger, then store
            t2t.insert(t2t_temp); 
            w2t.insert(w2t_temp);
            x2t.insert(x2t_temp);
            y2t.insert(y2t_temp);
            recordState(t2t_temp, TRIGGER2RECORD, 
               x2t_temp, y2t_temp, w2t_temp);
         } else {
//            cerr << "Error: incomplete stick 2 trigger information" 
//                 << ", flag value is: " << (int)s2tf << endl;
         }
         break;
   }
}



//////////////////////////////
//
// RadioDrum::s2ts -- "Stick 2 Trigger Status"
//    Keeps track of a compete set of stick 2 trigger positions.  
//    This function reset temporary values to new set of 
//    stick 2 trigger values.   Stores the timestamp and initilaizes
//    the s1tf status flag.
//

void RadioDrum::s2ts(long aTime) {
   whack2 = 1;                 // set the global state variable
   trigtime2 = aTime;          // set the global state variable 
   t2t_temp = aTime;           // store the internal temporary value
   s2tf = 0x00;                // initialize the flag to determine good set
   stick2trig();               // call the user state function
}



//////////////////////////////
//
// RadioDrum::s1pd -- "Stick 1 Position Data"
//    Stores position data in temporary storage
//    until a complete set of position data is received.
//

void RadioDrum::s1pd(int flag, uchar aValue) {
   switch (flag) {
      case DATA_X:
         xt1 = aValue;         // set the global state variable
         x1p_temp = aValue;    // set the internal temporary value
         s1pf += 1;            // set bit 1 (or create an error)
         break;
      case DATA_Y:
         yt1 = aValue;          // set the global state variable
         y1p_temp = aValue;     // set the internal temporary value
         s1pf += 4;             // set bit 3 (or create an error)
         break;
      case DATA_Z:
         zt1 = aValue;          // set the global state variable
         z1p_temp = aValue;     // set the internal temporary value
         s1pf += 16;            // set bit 5 (or create an error)
         if (s1pf == 21) {      // if complete temp position set then store
            x1p.insert(x1p_temp);
            y1p.insert(y1p_temp);
            z1p.insert(z1p_temp);
            recordState(t1p_temp, POSITION1RECORD, 
               x1p_temp, y1p_temp, z1p_temp);
         } else {
//            cerr << "Error: incomplete stick 1 position information"  
//                 << ", flag value is: " << (int)s1pf << endl;
         }
         break;
   }
}



//////////////////////////////
//
// RadioDrum::s1ps -- "Stick 1 Position Status"
//    Keeps track of a compete set of stick 1 positions.  
//    This function reset temporary values to new set of 
//    stick 1 position values.   Stores the timestamp and initilaizes
//    the s1pf status flag.
//

void RadioDrum::s1ps(long aTime) {
   t1p_temp = aTime;
   s1pf = 0x00;
}




//////////////////////////////
//
// RadioDrum::s2pd -- "Stick 2 Position Data"
//    Stores position data in temporary storage
//    until a complete set of position data is received.
//

void RadioDrum::s2pd(int flag, uchar aValue) {
   switch (flag) {
      case DATA_X:
         xt2 = aValue;          // set the global state variable
         x2p_temp = aValue;     // set the internal temporary variable
         s2pf += 1;             // set flag bit 1 (or create an error)
         break;
      case DATA_Y:
         yt2 = aValue;          // set the global state variable
         y2p_temp = aValue;     // set the internal temporary variable
         s2pf += 4;             // set flag bit 3 (or create an error)
         break;
      case DATA_Z:
         zt2 = aValue;          // set the global state variable
         z2p_temp = aValue;     // set the internal temporary variable
         s2pf += 16;            // set flag bit 5 (or create an error)
         if (s2pf == 21) {      // if complete temp position set then store
            x2p.insert(x2p_temp);
            y2p.insert(y2p_temp);
            z2p.insert(z2p_temp);
            recordState(t2p_temp, POSITION2RECORD, 
               x2p_temp, y2p_temp, z2p_temp);
         } else {
//            cerr << "Error: incomplete stick 2 position information" 
//                 << ", flag value is: " << (int)s2pf << endl;          
         }
         break;
   }
}



//////////////////////////////
//
// RadioDrum::s2ps -- "Stick 2 Position Status"
//    Keeps track of a compete set of stick 2 positions.  
//    This function reset temporary values to new set of 
//    stick 2 position values.   Stores the timestamp and initilaizes
//    the s1pf status flag.
//

void RadioDrum::s2ps(long aTime) {
   t2p_temp = aTime;
   s2pf = 0x00;
}



//////////////////////////////
//
// RadioDrum::writeCalibrationData -- writes calibration data
//	from calib array.  This function is called after
//	a successful transmission of calib data array from
//	the radio drum  or you can just write the current 
//	contents of the array as it is stored in the computer
//	(which may be different from the current state in the
//	radio drum.
//

void RadioDrum::writeCalibrationData(void) {
   // successful receipt of calibration, so write it to a file:
   fstream output(calibFilename, ios::out);
   if (!output.is_open()) {
      cerr << "Error: could not open " << calibFilename 
           << " to write data." << endl;
      // not a fatal error, however
   }

   output << "hit1      " << calib[0] << endl;
   output << "set1      " << calib[1] << endl;
   output << "hit2      " << calib[2] << endl;
   output << "set2      " << calib[3] << endl;
   output << "zero1x    " << calib[4] << endl;
   output << "zero1y    " << calib[5] << endl;
   output << "zero2x    " << calib[6] << endl;
   output << "zero2y    " << calib[7] << endl;
   output << "range     " << calib[8] << endl;
   output << "restpoint " << calib[9] << endl;
   for (int i=10; i<21; i++) {   // 22 elements in calib array
       output << calib[i] << endl;
   }
   output << "zcalib    " << calib[21] << endl;
}





